from onegov.core.csv import CSVFile
from onegov.core.errors import AmbiguousColumnsError
from onegov.core.errors import DuplicateColumnNamesError
from onegov.core.errors import EmptyFileError
from onegov.core.errors import EmptyLineInFileError
from onegov.core.errors import MissingColumnsError
from onegov.form.fields import UploadField
from onegov.form.validators import FileSizeLimit
from onegov.form.validators import WhitelistedMimeType
from onegov.wtfs import _


class CsvUploadField(UploadField):
    """ An upload field expecting CSV files. """

    def __init__(self, *args, **kwargs):
        self.encoding = kwargs.pop('encoding', None)
        self.dialect = kwargs.pop('dialect', None)
        self.expected_headers = kwargs.pop('expected_headers', None)
        self.rename_duplicate_column_names = kwargs.pop(
            'rename_duplicate_column_names', False
        )

        kwargs.setdefault('validators', [])
        kwargs['validators'].append(WhitelistedMimeType({'text/plain', }))
        kwargs['validators'].append(FileSizeLimit(10 * 1024 * 1024))

        kwargs.setdefault('render_kw', {})
        kwargs['render_kw']['force_simple'] = True

        super().__init__(*args, **kwargs)

    def post_validate(self, form, validation_stopped):
        super().post_validate(form, validation_stopped)
        if validation_stopped:
            return

        try:
            self.data = CSVFile(
                self.raw_data[0].file,
                encoding=self.encoding,
                dialect=self.dialect,
                expected_headers=self.expected_headers,
                rename_duplicate_column_names=(
                    self.rename_duplicate_column_names
                )
            )
            list(self.data.lines)
        except MissingColumnsError:
            raise ValueError(_("Some columns are missing."))
        except AmbiguousColumnsError:
            raise ValueError(_("Some column names are ambigous."))
        except DuplicateColumnNamesError:
            raise ValueError(_("Some column names appear twice."))
        except EmptyFileError:
            raise ValueError(_("The file is empty."))
        except EmptyLineInFileError:
            raise ValueError(_("The file contains an empty line."))
        except UnicodeDecodeError:
            raise ValueError(_("Invalid encoding."))
        except Exception:
            raise ValueError(_("Not a valid CSV file."))



from onegov.core.request import CoreRequest
from unittest.mock import patch


def test_views_notifications(client):
    client.login_admin()

    # Add a notification
    add = client.get('/notifications').click(href='/add')
    add.form['title'] = "Systemunterbruch"
    add.form['text'] = "Am 23. Februar 2017, finden Wartungsarbeiten statt."
    added = add.form.submit().follow()
    assert "Benachrichtigung hinzugefügt." in added
    assert "Systemunterbruch" in added

    # Dashboard
    view = client.get('/')
    assert "Systemunterbruch" in view
    assert "Wartungsarbeiten" in view

    # View the notification
    view = client.get('/notifications').click("Systemunterbruch")
    assert "Wartungsarbeiten" in view

    # Edit the notification
    edit = view.click("Bearbeiten")
    edit.form['title'] = "Wartungsarbeiten"
    edit.form['text'] = "24. Februar 2017"
    edited = edit.form.submit().follow()
    assert "Benachrichtigung geändert." in edited
    assert "Wartungsarbeiten" in edited

    # Delete the notification
    deleted = client.get('/notifications')\
        .click("Wartungsarbeiten").click("Löschen")
    assert deleted.status_int == 200
    assert "Wartungsarbeiten" not in client.get('/notifications')
    assert "Wartungsarbeiten" not in client.get('/').follow()


@patch.object(CoreRequest, 'assert_valid_csrf_token')
def test_views_notifications_permissions(mock_method, client):
    client.login_admin()

    add = client.get('/notifications').click(href='/add')
    add.form['title'] = "Systemunterbruch"
    add.form['text'] = "Am 23. Februar 2017, finden Wartungsarbeiten statt."
    assert "Benachrichtigung hinzugefügt." in add.form.submit().follow()
    id = client.get('/notifications')\
        .click("Systemunterbruch").request.url.split('/')[-1]

    client.logout()

    client.get('/notifications', status=403)
    client.get('/notifications/add', status=403)
    client.get(f'/notification/{id}', status=403)
    client.get(f'/notification/{id}/edit', status=403)
    client.delete(f'/notification/{id}', status=403)

    client.login_member()
    client.get('/notifications', status=403)
    client.get('/notifications/add', status=403)
    client.get(f'/notification/{id}')
    client.get(f'/notification/{id}/edit', status=403)
    client.delete(f'/notification/{id}', status=403)
    client.logout()

    client.login_editor()
    client.get('/notifications', status=403)
    client.get('/notifications/add', status=403)
    client.get(f'/notification/{id}')
    client.get(f'/notification/{id}/edit', status=403)
    client.delete(f'/notification/{id}', status=403)
    client.logout()

    client.login_admin()
    client.get('/notifications')
    client.get('/notifications/add')
    client.get(f'/notification/{id}')
    client.get(f'/notification/{id}/edit')
    client.delete(f'/notification/{id}')
    client.logout()



from onegov.core.request import CoreRequest
from unittest.mock import patch
from webtest.forms import Upload


def test_views_municipality(client):
    client.login_admin()

    # Add a municipality
    add = client.get('/municipalities').click(href='/add')
    add.form['name'] = "Adlikon"
    add.form['bfs_number'] = '21'
    add.form['payment_type'] = 'normal'
    added = add.form.submit().follow()
    assert "hinzugefügt." in added
    assert "Adlikon" in added

    # View the municipality
    view = client.get('/municipalities').click("Adlikon")
    assert "Adlikon" in view
    assert "21" in view
    assert "Normal" in view
    assert "7.00" in view

    # Edit the municipality
    edit = view.click("Bearbeiten")
    edit.form['name'] = "Aesch"
    edit.form['bfs_number'] = '241'
    edit.form['payment_type'] = 'spezial'
    edit.form['address_supplement'] = "Zusatz"
    edit.form['gpn_number'] = "12321"
    edited = edit.form.submit().follow()
    assert "geändert." in edited
    assert "Aesch" in edited
    view = edited.click("Aesch")
    assert "Spezial" in view
    assert "8.50" in view
    assert "12321" in view

    # Upload some dates
    upload = client.get('/municipalities').click("Daten importieren")
    upload.form['file'] = Upload(
        'test.csv',
        "Gemeinde-Nr,Vordefinierte Termine\n241,12.2.2015".encode('utf-8'),
        'text/csv'
    )
    uploaded = upload.form.submit().follow()
    assert "Gemeindedaten importiert." in uploaded
    assert "12.02.2015" in client.get('/municipalities').click("Aesch")

    # Delete some dates
    clear = client.get('/municipalities').click("Aesch")\
        .click("Abholtermine löschen")
    cleared = clear.form.submit().follow()
    assert "Abholtermine gelöscht." in cleared
    assert "12.02.2015" not in client.get('/municipalities').click("Aesch")

    # Delete the municipality
    deleted = client.get('/municipalities').click("Aesch").click("Löschen")
    assert deleted.status_int == 200
    assert "Aesch" not in client.get('/municipalities')


@patch.object(CoreRequest, 'assert_valid_csrf_token')
def test_views_municipality_permissions(mock_method, client):
    client.login_admin()

    add = client.get('/municipalities').click(href='/add')
    add.form['name'] = "Adlikon"
    add.form['bfs_number'] = '21'
    add.form['payment_type'] = 'normal'
    assert "hinzugefügt." in add.form.submit().follow()
    id = client.get('/municipalities')\
        .click("Adlikon").request.url.split('/')[-1]

    client.logout()

    urls = [
        '/municipalities',
        '/municipalities/add',
        f'/municipality/{id}',
        f'/municipality/{id}/edit'
    ]

    for url in urls:
        client.get(url, status=403)
    client.delete(f'/municipality/{id}', status=403)

    client.login_member()
    for url in urls:
        client.get(url, status=403)
    client.delete(f'/municipality/{id}', status=403)
    client.logout()

    client.login_editor()
    for url in urls:
        client.get(url, status=403)
    client.delete(f'/municipality/{id}', status=403)
    client.logout()

    client.login_admin()
    for url in urls:
        client.get(url)
    client.delete(f'/municipality/{id}')
    client.logout()



from onegov.core.orm.func import unaccent
from onegov.form import Form
from onegov.wtfs import _
from onegov.wtfs.fields import MunicipalityDataUploadField
from onegov.wtfs.models import Municipality
from onegov.wtfs.models import PaymentType
from onegov.wtfs.models import PickupDate
from wtforms import IntegerField
from wtforms import RadioField
from wtforms import SelectField
from wtforms import StringField
from wtforms.fields.html5 import DateField
from wtforms.validators import DataRequired
from wtforms.validators import InputRequired
from wtforms.validators import Optional


class MunicipalityForm(Form):

    name = StringField(
        label=_("Name"),
        validators=[
            InputRequired()
        ]
    )

    bfs_number = IntegerField(
        label=_("BFS number"),
        validators=[
            InputRequired()
        ]
    )

    address_supplement = StringField(
        label=_("Address supplement"),
    )

    gpn_number = IntegerField(
        label=_("GPN number"),
        validators=[
            Optional()
        ]
    )

    payment_type = RadioField(
        label=_("Payment type"),
        validators=[InputRequired()]
    )

    def on_request(self):
        query = self.request.session.query(PaymentType.name)
        self.payment_type.choices = [
            (r.name, r.name.capitalize()) for r in query
        ]

    def update_model(self, model):
        model.name = self.name.data
        model.bfs_number = self.bfs_number.data
        model.address_supplement = self.address_supplement.data
        model.gpn_number = self.gpn_number.data
        model.payment_type = self.payment_type.data

    def apply_model(self, model):
        self.name.data = model.name
        self.bfs_number.data = model.bfs_number
        self.address_supplement.data = model.address_supplement
        self.gpn_number.data = model.gpn_number
        self.payment_type.data = model.payment_type


class ImportMunicipalityDataForm(Form):

    file = MunicipalityDataUploadField(
        label=_("File"),
        validators=[
            DataRequired()
        ]
    )

    def update_model(self, model):
        model.import_data(self.file.data)


class DeleteMunicipalityDatesForm(Form):

    start = DateField(
        label=_("Start date"),
        validators=[
            DataRequired()
        ]
    )

    end = DateField(
        label=_("End date"),
        validators=[
            DataRequired()
        ]
    )

    def update_model(self, model):
        dates = model.pickup_dates
        dates = dates.filter(PickupDate.date >= self.start.data)
        dates = dates.filter(PickupDate.date <= self.end.data)
        for date in dates:
            self.request.session.delete(date)

    def apply_model(self, model):
        start = model.pickup_dates.first()
        self.start.data = start.date if start else None

        end = model.pickup_dates.order_by(None)
        end = end.order_by(PickupDate.date.desc()).first()
        self.end.data = end.date if end else None


class MunicipalityIdSelectionForm(Form):

    municipality_id = SelectField(
        label=_("Municipality"),
        choices=[],
        validators=[
            InputRequired()
        ]
    )

    def on_request(self):
        query = self.request.session.query(
            Municipality.id.label('id'),
            Municipality.name.label('name')
        )
        query = query.order_by(unaccent(Municipality.name))
        self.municipality_id.choices = [(r.id.hex, r.name) for r in query]

    @property
    def municipality(self):
        if self.municipality_id.data:
            query = self.request.session.query(Municipality)
            query = query.filter(Municipality.id == self.municipality_id.data)
            return query.first()



from onegov.core.orm.func import unaccent
from onegov.form import Form
from onegov.form.fields import PreviewField
from onegov.wtfs import _
from onegov.wtfs.models import Municipality
from wtforms import SelectField
from wtforms import StringField
from wtforms.fields.html5 import DateField
from wtforms.validators import InputRequired


class CreateInvoicesForm(Form):

    from_date = DateField(
        label=_("From date"),
        validators=[
            InputRequired(),
        ]
    )

    to_date = DateField(
        label=_("From date"),
        validators=[
            InputRequired(),
        ]
    )

    cs2_user = StringField(
        label=_("CS2 User"),
        validators=[
            InputRequired(),
        ]
    )

    subject = StringField(
        label=_("Subject"),
        validators=[
            InputRequired(),
        ]
    )

    municipality_id = SelectField(
        label=_("Municipality"),
        choices=[],
        validators=[InputRequired()]
    )

    price_per_quantity = PreviewField(
        label=_("Price per quantity"),
        fields=('municipality_id',),
        events=('change',),
        url=lambda meta: meta.request.link(
            meta.request.app.principal,
            name='price-per-quantity'
        )
    )

    accounting_unit = StringField(
        label=_("Accounting unit"),
        validators=[
            InputRequired(),
        ]
    )

    revenue_account = StringField(
        label=_("Revenue account"),
        validators=[
            InputRequired(),
        ]
    )

    def on_request(self):
        query = self.request.session.query(
            Municipality.id.label('id'),
            Municipality.name.label('name'),
            Municipality.meta['bfs_number'].label('bfs_number'),
        )
        query = query.order_by(unaccent(Municipality.name))
        self.municipality_id.choices = [('-', '-')]
        self.municipality_id.choices += [
            (r.id.hex, f"{r.name} ({r.bfs_number})") for r in query
        ]

    def update_model(self, model):
        model.from_date = self.from_date.data
        model.to_date = self.to_date.data
        model.cs2_user = self.cs2_user.data
        model.subject = self.subject.data
        model.municipality_id = None
        if self.municipality_id.data != '-':
            model.municipality_id = self.municipality_id.data
        model.accounting_unit = self.accounting_unit.data
        model.revenue_account = self.revenue_account.data



from datetime import date
from morepath import Identity
from onegov.core.security import Public
from onegov.user import User
from onegov.user import UserCollection
from onegov.user import UserGroup
from onegov.user import UserGroupCollection
from onegov.wtfs.collections import MunicipalityCollection
from onegov.wtfs.collections import PaymentTypeCollection
from onegov.wtfs.collections import ScanJobCollection
from onegov.wtfs.models import DailyList
from onegov.wtfs.models import DailyListBoxes
from onegov.wtfs.models import DailyListBoxesAndForms
from onegov.wtfs.models import Invoice
from onegov.wtfs.models import Municipality
from onegov.wtfs.models import Notification
from onegov.wtfs.models import PaymentType
from onegov.wtfs.models import PickupDate
from onegov.wtfs.models import Report
from onegov.wtfs.models import ReportBoxes
from onegov.wtfs.models import ReportBoxesAndForms
from onegov.wtfs.models import ReportFormsByMunicipality
from onegov.wtfs.models import ScanJob
from onegov.wtfs.security import AddModel
from onegov.wtfs.security import AddModelUnrestricted
from onegov.wtfs.security import DeleteModel
from onegov.wtfs.security import EditModel
from onegov.wtfs.security import EditModelUnrestricted
from onegov.wtfs.security import ViewModel
from onegov.wtfs.security import ViewModelUnrestricted
from uuid import uuid4


def test_permissions(wtfs_app, wtfs_password):
    session = wtfs_app.session()

    # Remove existing users and group
    session.query(User).filter_by(realname='Editor').one().group_id = None
    session.query(User).filter_by(realname='Member').one().group_id = None
    session.query(Municipality).delete()

    # Add new
    municipality_a_id = uuid4()
    session.add(Municipality(
        id=municipality_a_id,
        name='Municipality A',
        bfs_number=10,
    ))
    session.add(PickupDate(
        date=date.today(),
        municipality_id=municipality_a_id,
    ))
    session.add(ScanJob(
        type='normal',
        municipality_id=municipality_a_id,
        dispatch_date=date(2019, 1, 1))
    )
    session.add(User(
        realname='Admin A',
        username='admin-a@example.org',
        password_hash=wtfs_password,
        role='admin',
        group_id=municipality_a_id
    ))
    session.add(User(
        realname='Editor A',
        username='editor-a@example.org',
        password_hash=wtfs_password,
        role='editor',
        group_id=municipality_a_id
    ))
    session.add(User(
        realname='Member A',
        username='member-a@example.org',
        password_hash=wtfs_password,
        role='member',
        group_id=municipality_a_id
    ))

    municipality_b_id = uuid4()
    session.add(Municipality(
        id=municipality_b_id,
        name='Municipality B',
        bfs_number=20,
    ))
    session.add(User(
        realname='Admin B',
        username='admin-b@example.org',
        password_hash=wtfs_password,
        role='admin',
        group_id=municipality_b_id
    ))
    session.add(User(
        realname='Editor B',
        username='editor-b@example.org',
        password_hash=wtfs_password,
        role='editor',
        group_id=municipality_b_id
    ))
    session.add(User(
        realname='Member B',
        username='member-b@example.org',
        password_hash=wtfs_password,
        role='member',
        group_id=municipality_b_id
    ))

    query = session.query
    admin = query(User).filter_by(realname='Admin').one()
    admin_a = query(User).filter_by(realname='Admin A').one()
    admin_b = query(User).filter_by(realname='Admin B').one()
    editor = query(User).filter_by(realname='Editor').one()
    editor_a = query(User).filter_by(realname='Editor A').one()
    editor_b = query(User).filter_by(realname='Editor B').one()
    member = query(User).filter_by(realname='Member').one()
    member_a = query(User).filter_by(realname='Member A').one()
    member_b = query(User).filter_by(realname='Member B').one()
    group = UserGroup()
    group_a = query(UserGroup).filter_by(name='Municipality A').one()
    group_b = query(UserGroup).filter_by(name='Municipality B').one()
    municipality = Municipality()
    municipality_a = query(Municipality).filter_by(name='Municipality A').one()
    municipality_b = query(Municipality).filter_by(name='Municipality B').one()
    scan_job = query(ScanJob).one()

    def permits(user, model, permission):
        return wtfs_app._permits(
            Identity(
                userid=user.username,
                groupid=user.group_id.hex if user.group_id else '',
                role=user.role,
                application_id=wtfs_app.application_id
            ),
            model,
            permission
        )

    # General
    model = object()
    for user in (admin, admin_a, admin_b):
        assert permits(user, model, Public)
        assert permits(user, model, AddModel)
        assert permits(user, model, AddModelUnrestricted)
        assert permits(user, model, EditModel)
        assert permits(user, model, EditModelUnrestricted)
        assert permits(user, model, DeleteModel)
        assert permits(user, model, ViewModel)
        assert permits(user, model, ViewModelUnrestricted)
    for user in (editor, editor_a, editor_b, member, member_a, member_b):
        assert permits(user, model, Public)
        assert not permits(user, model, AddModel)
        assert not permits(user, model, AddModelUnrestricted)
        assert not permits(user, model, EditModel)
        assert not permits(user, model, EditModelUnrestricted)
        assert not permits(user, model, DeleteModel)
        assert not permits(user, model, ViewModel)
        assert not permits(user, model, ViewModelUnrestricted)

    # UserGroupCollection / MunicipalityCollection
    # MunicipalityCollection
    for model in (
        MunicipalityCollection(session),
        UserGroupCollection(session)
    ):
        for user in (admin, admin_a, admin_b):
            assert permits(user, model, Public)
            assert permits(user, model, AddModel)
            assert permits(user, model, AddModelUnrestricted)
            assert permits(user, model, EditModel)
            assert permits(user, model, EditModelUnrestricted)
            assert permits(user, model, DeleteModel)
            assert permits(user, model, ViewModel)
            assert permits(user, model, ViewModelUnrestricted)
        for user in (editor, editor_a, editor_b, member, member_a, member_b):
            assert permits(user, model, Public)
            assert not permits(user, model, AddModel)
            assert not permits(user, model, AddModelUnrestricted)
            assert not permits(user, model, EditModel)
            assert not permits(user, model, EditModelUnrestricted)
            assert not permits(user, model, DeleteModel)
            assert not permits(user, model, ViewModel)
            assert not permits(user, model, ViewModelUnrestricted)

    # UserGroup / Municipality
    for user in (admin, admin_a, admin_b):
        for model in (group_a, group_b, municipality_a, municipality_b):
            assert permits(user, model, Public)
            assert permits(user, model, AddModel)
            assert permits(user, model, AddModelUnrestricted)
            assert permits(user, model, EditModel)
            assert permits(user, model, EditModelUnrestricted)
            assert not permits(user, model, DeleteModel)
            assert permits(user, model, ViewModel)
            assert permits(user, model, ViewModelUnrestricted)
        for model in (group, municipality):
            assert permits(user, model, Public)
            assert permits(user, model, AddModel)
            assert permits(user, model, AddModelUnrestricted)
            assert permits(user, model, EditModel)
            assert permits(user, model, EditModelUnrestricted)
            assert permits(user, model, DeleteModel)
            assert permits(user, model, ViewModel)
            assert permits(user, model, ViewModelUnrestricted)
    for user in (editor, editor_a, editor_b, member, member_a, member_b):
        for model in (
            group_a, group_b, group,
            municipality_a, municipality_b, municipality
        ):
            assert permits(user, model, Public)
            assert not permits(user, model, AddModel)
            assert not permits(user, model, AddModelUnrestricted)
            assert not permits(user, model, EditModel)
            assert not permits(user, model, EditModelUnrestricted)
            assert not permits(user, model, DeleteModel)
            assert not permits(user, model, ViewModel)
            assert not permits(user, model, ViewModelUnrestricted)

    # UserCollection
    model = UserCollection(session)
    for user in (admin, admin_a, admin_b):
        assert permits(user, model, Public)
        assert permits(user, model, AddModel)
        assert permits(user, model, AddModelUnrestricted)
        assert permits(user, model, EditModel)
        assert permits(user, model, EditModelUnrestricted)
        assert permits(user, model, DeleteModel)
        assert permits(user, model, ViewModel)
        assert permits(user, model, ViewModelUnrestricted)
    for user in (editor, editor_a, editor_b):
        assert permits(user, model, Public)
        assert permits(user, model, AddModel)
        assert not permits(user, model, AddModelUnrestricted)
        assert not permits(user, model, EditModel)
        assert not permits(user, model, EditModelUnrestricted)
        assert not permits(user, model, DeleteModel)
        assert permits(user, model, ViewModel)
        assert not permits(user, model, ViewModelUnrestricted)
    for user in (member, member_a, member_b):
        assert permits(user, model, Public)
        assert not permits(user, model, AddModel)
        assert not permits(user, model, AddModelUnrestricted)
        assert not permits(user, model, EditModel)
        assert not permits(user, model, EditModelUnrestricted)
        assert not permits(user, model, DeleteModel)
        assert not permits(user, model, ViewModel)
        assert not permits(user, model, ViewModelUnrestricted)

    # User
    for user in (admin, admin_a, admin_b):
        for model in (admin, admin_a, admin_b):
            assert permits(user, model, Public)
            assert permits(user, model, AddModel)
            assert permits(user, model, AddModelUnrestricted)
            assert permits(user, model, EditModel)
            assert permits(user, model, EditModelUnrestricted)
            if user == model:
                assert not permits(user, model, DeleteModel)
            else:
                assert permits(user, model, DeleteModel)
            assert permits(user, model, ViewModel)
            assert permits(user, model, ViewModelUnrestricted)
        for model in (editor, editor_a, editor_b, member, member_a, member_b):
            assert permits(user, model, Public)
            assert permits(user, model, AddModel)
            assert permits(user, model, AddModelUnrestricted)
            assert permits(user, model, EditModel)
            assert permits(user, model, EditModelUnrestricted)
            assert permits(user, model, DeleteModel)
            assert permits(user, model, ViewModel)
            assert permits(user, model, ViewModelUnrestricted)
    for user in (editor, ):
        for model in (admin, admin_a, admin_b, editor, editor_a, editor_b,
                      member, member_a, member_b):
            assert permits(user, model, Public)
            assert not permits(user, model, AddModel)
            assert not permits(user, model, AddModelUnrestricted)
            assert not permits(user, model, EditModel)
            assert not permits(user, model, EditModelUnrestricted)
            assert not permits(user, model, DeleteModel)
            assert not permits(user, model, ViewModel)
            assert not permits(user, model, ViewModelUnrestricted)
    for user in (editor_a, ):
        for model in (admin, admin_a, admin_b, editor, editor_a, editor_b,
                      member, member_b):
            assert permits(user, model, Public)
            assert not permits(user, model, AddModel)
            assert not permits(user, model, AddModelUnrestricted)
            assert not permits(user, model, EditModel)
            assert not permits(user, model, EditModelUnrestricted)
            assert not permits(user, model, DeleteModel)
            assert not permits(user, model, ViewModel)
            assert not permits(user, model, ViewModelUnrestricted)
        for model in (member_a,):
            assert permits(user, model, Public)
            assert not permits(user, model, AddModel)
            assert not permits(user, model, AddModelUnrestricted)
            assert permits(user, model, EditModel)
            assert not permits(user, model, EditModelUnrestricted)
            assert permits(user, model, DeleteModel)
            assert permits(user, model, ViewModel)
            assert not permits(user, model, ViewModelUnrestricted)
    for user in (editor_b, ):
        for model in (admin, admin_a, admin_b, editor, editor_a, editor_b,
                      member, member_a):
            assert permits(user, model, Public)
            assert not permits(user, model, AddModel)
            assert not permits(user, model, AddModelUnrestricted)
            assert not permits(user, model, EditModel)
            assert not permits(user, model, EditModelUnrestricted)
            assert not permits(user, model, DeleteModel)
            assert not permits(user, model, ViewModel)
            assert not permits(user, model, ViewModelUnrestricted)
        for model in (member_b,):
            assert permits(user, model, Public)
            assert not permits(user, model, AddModel)
            assert not permits(user, model, AddModelUnrestricted)
            assert permits(user, model, EditModel)
            assert not permits(user, model, EditModelUnrestricted)
            assert permits(user, model, DeleteModel)
            assert permits(user, model, ViewModel)
            assert not permits(user, model, ViewModelUnrestricted)
    for user in (member, member_a, member_b):
        for model in (admin, admin_a, admin_b, editor, editor_a, editor_b,
                      member, member_a, member_b):
            assert permits(user, model, Public)
            assert not permits(user, model, AddModel)
            assert not permits(user, model, AddModelUnrestricted)
            assert not permits(user, model, EditModel)
            assert not permits(user, model, EditModelUnrestricted)
            assert not permits(user, model, DeleteModel)
            assert not permits(user, model, ViewModel)
            assert not permits(user, model, ViewModelUnrestricted)

    # ScanJobCollection
    model = ScanJobCollection(session)
    for user in (admin, admin_a, admin_b):
        assert permits(user, model, Public)
        assert permits(user, model, AddModel)
        assert permits(user, model, AddModelUnrestricted)
        assert permits(user, model, EditModel)
        assert permits(user, model, EditModelUnrestricted)
        assert permits(user, model, DeleteModel)
        assert permits(user, model, ViewModel)
        assert permits(user, model, ViewModelUnrestricted)
    for user in (editor_a, editor_b, member_a, member_b):
        assert permits(user, model, Public)
        assert permits(user, model, AddModel)
        assert not permits(user, model, AddModelUnrestricted)
        assert not permits(user, model, EditModel)
        assert not permits(user, model, EditModelUnrestricted)
        assert not permits(user, model, DeleteModel)
        assert permits(user, model, ViewModel)
        assert not permits(user, model, ViewModelUnrestricted)
    for user in (editor, member):
        assert permits(user, model, Public)
        assert not permits(user, model, AddModel)
        assert not permits(user, model, AddModelUnrestricted)
        assert not permits(user, model, EditModel)
        assert not permits(user, model, EditModelUnrestricted)
        assert not permits(user, model, DeleteModel)
        assert not permits(user, model, ViewModel)
        assert not permits(user, model, ViewModelUnrestricted)

    # ScanJob
    model = scan_job
    for user in (admin, admin_a, admin_b):
        assert permits(user, model, Public)
        assert permits(user, model, AddModel)
        assert permits(user, model, AddModelUnrestricted)
        assert permits(user, model, EditModel)
        assert permits(user, model, EditModelUnrestricted)
        assert permits(user, model, DeleteModel)
        assert permits(user, model, ViewModel)
        assert permits(user, model, ViewModelUnrestricted)
    for user in (editor, editor_b, member, member_b):
        assert permits(user, model, Public)
        assert not permits(user, model, AddModel)
        assert not permits(user, model, AddModelUnrestricted)
        assert not permits(user, model, EditModel)
        assert not permits(user, model, EditModelUnrestricted)
        assert not permits(user, model, DeleteModel)
        assert not permits(user, model, ViewModel)
        assert not permits(user, model, ViewModelUnrestricted)
    for user in (editor_a, member_a):
        assert permits(user, model, Public)
        assert not permits(user, model, AddModel)
        assert not permits(user, model, AddModelUnrestricted)
        assert permits(user, model, EditModel)
        assert not permits(user, model, EditModelUnrestricted)
        assert not permits(user, model, DeleteModel)
        assert permits(user, model, ViewModel)
        assert not permits(user, model, ViewModelUnrestricted)

    # DailyList
    for model in (DailyList(), DailyListBoxes(session)):
        for user in (admin, admin_a, admin_b):
            assert permits(user, model, Public)
            assert permits(user, model, AddModel)
            assert permits(user, model, AddModelUnrestricted)
            assert permits(user, model, EditModel)
            assert permits(user, model, EditModelUnrestricted)
            assert permits(user, model, DeleteModel)
            assert permits(user, model, ViewModel)
            assert permits(user, model, ViewModelUnrestricted)
        for user in (editor, editor_a, editor_b, member_a, member_b):
            assert permits(user, model, Public)
            assert not permits(user, model, AddModel)
            assert not permits(user, model, AddModelUnrestricted)
            assert not permits(user, model, EditModel)
            assert not permits(user, model, EditModelUnrestricted)
            assert not permits(user, model, DeleteModel)
            assert not permits(user, model, ViewModel)
            assert not permits(user, model, ViewModelUnrestricted)
        for user in (member,):
            assert permits(user, model, Public)
            assert not permits(user, model, AddModel)
            assert not permits(user, model, AddModelUnrestricted)
            assert not permits(user, model, EditModel)
            assert not permits(user, model, EditModelUnrestricted)
            assert not permits(user, model, DeleteModel)
            assert permits(user, model, ViewModel)
            assert not permits(user, model, ViewModelUnrestricted)
    for model in (DailyListBoxesAndForms(session), ):
        for user in (admin, admin_a, admin_b):
            assert permits(user, model, Public)
            assert permits(user, model, AddModel)
            assert permits(user, model, AddModelUnrestricted)
            assert permits(user, model, EditModel)
            assert permits(user, model, EditModelUnrestricted)
            assert permits(user, model, DeleteModel)
            assert permits(user, model, ViewModel)
            assert permits(user, model, ViewModelUnrestricted)
        for user in (editor, editor_a, editor_b, member, member_a, member_b):
            assert permits(user, model, Public)
            assert not permits(user, model, AddModel)
            assert not permits(user, model, AddModelUnrestricted)
            assert not permits(user, model, EditModel)
            assert not permits(user, model, EditModelUnrestricted)
            assert not permits(user, model, DeleteModel)
            assert not permits(user, model, ViewModel)
            assert not permits(user, model, ViewModelUnrestricted)

    # Report
    for model in (
        Report(session),
        ReportBoxes(session),
        ReportBoxesAndForms(session),
        ReportFormsByMunicipality(session)
    ):
        for user in (admin, admin_a, admin_b):
            assert permits(user, model, Public)
            assert permits(user, model, AddModel)
            assert permits(user, model, AddModelUnrestricted)
            assert permits(user, model, EditModel)
            assert permits(user, model, EditModelUnrestricted)
            assert permits(user, model, DeleteModel)
            assert permits(user, model, ViewModel)
            assert permits(user, model, ViewModelUnrestricted)
        for user in (editor, editor_a, editor_b, member, member_a, member_b):
            assert permits(user, model, Public)
            assert not permits(user, model, AddModel)
            assert not permits(user, model, AddModelUnrestricted)
            assert not permits(user, model, EditModel)
            assert not permits(user, model, EditModelUnrestricted)
            assert not permits(user, model, DeleteModel)
            assert not permits(user, model, ViewModel)
            assert not permits(user, model, ViewModelUnrestricted)

    # Notification
    model = Notification()
    for user in (admin, admin_a, admin_b):
        assert permits(user, model, Public)
        assert permits(user, model, AddModel)
        assert permits(user, model, AddModelUnrestricted)
        assert permits(user, model, EditModel)
        assert permits(user, model, EditModelUnrestricted)
        assert permits(user, model, DeleteModel)
        assert permits(user, model, ViewModel)
        assert permits(user, model, ViewModelUnrestricted)
    for user in (editor, editor_a, editor_b, member, member_a, member_b):
        assert permits(user, model, Public)
        assert not permits(user, model, AddModel)
        assert not permits(user, model, AddModelUnrestricted)
        assert not permits(user, model, EditModel)
        assert not permits(user, model, EditModelUnrestricted)
        assert not permits(user, model, DeleteModel)
        assert permits(user, model, ViewModel)
        assert not permits(user, model, ViewModelUnrestricted)

    # Invoice
    model = Invoice(session)
    for user in (admin, admin_a, admin_b):
        assert permits(user, model, Public)
        assert permits(user, model, AddModel)
        assert permits(user, model, AddModelUnrestricted)
        assert permits(user, model, EditModel)
        assert permits(user, model, EditModelUnrestricted)
        assert permits(user, model, DeleteModel)
        assert permits(user, model, ViewModel)
        assert permits(user, model, ViewModelUnrestricted)
    for user in (editor, editor_a, editor_b, member, member_a, member_b):
        assert permits(user, model, Public)
        assert not permits(user, model, AddModel)
        assert not permits(user, model, AddModelUnrestricted)
        assert not permits(user, model, EditModel)
        assert not permits(user, model, EditModelUnrestricted)
        assert not permits(user, model, DeleteModel)
        assert not permits(user, model, ViewModel)
        assert not permits(user, model, ViewModelUnrestricted)

    # PaymentType
    model = PaymentType()
    for user in (admin, admin_a, admin_b):
        assert permits(user, model, Public)
        assert permits(user, model, AddModel)
        assert permits(user, model, AddModelUnrestricted)
        assert permits(user, model, EditModel)
        assert permits(user, model, EditModelUnrestricted)
        assert permits(user, model, DeleteModel)
        assert permits(user, model, ViewModel)
        assert permits(user, model, ViewModelUnrestricted)
    for user in (editor, editor_a, editor_b, member, member_a, member_b):
        assert permits(user, model, Public)
        assert not permits(user, model, AddModel)
        assert not permits(user, model, AddModelUnrestricted)
        assert not permits(user, model, EditModel)
        assert not permits(user, model, EditModelUnrestricted)
        assert not permits(user, model, DeleteModel)
        assert not permits(user, model, ViewModel)
        assert not permits(user, model, ViewModelUnrestricted)

    # PaymentTypeCollection
    model = PaymentTypeCollection(session)
    for user in (admin, admin_a, admin_b):
        assert permits(user, model, Public)
        assert permits(user, model, AddModel)
        assert permits(user, model, AddModelUnrestricted)
        assert permits(user, model, EditModel)
        assert permits(user, model, EditModelUnrestricted)
        assert permits(user, model, DeleteModel)
        assert permits(user, model, ViewModel)
        assert permits(user, model, ViewModelUnrestricted)
    for user in (editor, editor_a, editor_b, member, member_a, member_b):
        assert permits(user, model, Public)
        assert not permits(user, model, AddModel)
        assert not permits(user, model, AddModelUnrestricted)
        assert not permits(user, model, EditModel)
        assert not permits(user, model, EditModelUnrestricted)
        assert not permits(user, model, DeleteModel)
        assert not permits(user, model, ViewModel)
        assert not permits(user, model, ViewModelUnrestricted)



from cached_property import cached_property
from datetime import date
from onegov.core.orm.func import unaccent
from onegov.wtfs.models.municipality import Municipality
from onegov.wtfs.models.scan_job import ScanJob
from sqlalchemy import func
from sqlalchemy import Integer
from sqlalchemy.sql.expression import literal_column


def sum(table, attribute):
    result = func.coalesce(func.sum(getattr(table, attribute)), 0)
    result = func.cast(result, Integer)
    return result.label(attribute)


def zero(attribute):
    return literal_column('0').label(attribute)


class Report(object):
    """ The base class for the reports.

    Aggregates the ``columns_dispatch`` on the dispatch date and
    ``columns_return`` on the return date.

    Allows to filter by date range and scan job type.

    """

    def __init__(
        self, session, start=None, end=None, type=None, municipality_id=None
    ):
        self.session = session
        self.start = start or date.today()
        self.end = end or date.today()
        self.type = type
        self.municipality_id = municipality_id

    @cached_property
    def municipality_name(self):
        if self.municipality_id:
            query = self.session.query(Municipality.name)
            query = query.filter_by(id=self.municipality_id)
            return query.scalar()

    @cached_property
    def columns_dispatch(self):
        return []

    @cached_property
    def columns_return(self):
        return []

    @cached_property
    def columns(self):
        return self.columns_dispatch + self.columns_return

    def query(self):
        # aggregate on dispatch date
        query_in = self.session.query(ScanJob).join(Municipality)
        query_in = query_in.with_entities(
            Municipality.name.label('name'),
            Municipality.meta['bfs_number'].label('bfs_number'),
            *[sum(ScanJob, column) for column in self.columns_dispatch],
            *[zero(column) for column in self.columns_return],
        )
        query_in = query_in.filter(
            ScanJob.dispatch_date >= self.start,
            ScanJob.dispatch_date <= self.end
        )
        if self.type in ('normal', 'express'):
            query_in = query_in.filter(ScanJob.type == self.type)
        if self.municipality_id:
            query_in = query_in.filter(
                Municipality.id == self.municipality_id
            )
        query_in = query_in.group_by(
            Municipality.name,
            Municipality.meta['bfs_number'].label('bfs_number')
        )

        # aggregate on return date
        query_out = self.session.query(ScanJob).join(Municipality)
        query_out = query_out.with_entities(
            Municipality.name.label('name'),
            Municipality.meta['bfs_number'].label('bfs_number'),
            *[zero(column) for column in self.columns_dispatch],
            *[sum(ScanJob, column) for column in self.columns_return],
        )
        query_out = query_out.filter(
            ScanJob.return_date >= self.start,
            ScanJob.return_date <= self.end
        )
        if self.type in ('normal', 'express'):
            query_out = query_out.filter(ScanJob.type == self.type)
        if self.municipality_id:
            query_out = query_out.filter(
                Municipality.id == self.municipality_id
            )
        query_out = query_out.group_by(
            Municipality.name,
            Municipality.meta['bfs_number'].label('bfs_number')
        )

        # join
        union = query_in.union_all(query_out).subquery('union')
        query = self.session.query(
            union.c.name,
            union.c.bfs_number,
            *[sum(union.c, column) for column in self.columns]
        )
        query = query.group_by(union.c.name, union.c.bfs_number)
        query = query.order_by(unaccent(union.c.name))
        return query

    def total(self):
        subquery = self.query().subquery()
        query = self.session.query(
            *[sum(subquery.c, column) for column in self.columns],
        )
        return query.one()


class ReportBoxes(Report):
    """ A report containing all boxes from the municipalities of normal scan
    jobs. """

    def __init__(self, session, start=None, end=None):
        super().__init__(session, start, end, ['normal'])

    @cached_property
    def columns_dispatch(self):
        return [
            'dispatch_boxes',
            'dispatch_cantonal_tax_office',
            'dispatch_cantonal_scan_center',
            'return_boxes'
        ]


class ReportBoxesAndForms(Report):
    """ A report containing all boxes, tax forms and single documents. """

    @cached_property
    def columns_dispatch(self):
        return [
            'return_scanned_tax_forms_older',
            'return_scanned_tax_forms_last_year',
            'return_scanned_tax_forms_current_year',
            'return_scanned_tax_forms',
            'return_scanned_single_documents',
            'return_boxes'
        ]


class ReportFormsByMunicipality(Report):
    """ A report containing all tax forms of a single municipality. """

    @cached_property
    def columns_dispatch(self):
        return [
            'return_scanned_tax_forms_older',
            'return_scanned_tax_forms_last_year',
            'return_scanned_tax_forms_current_year',
            'return_scanned_tax_forms',
        ]


class ReportBoxesAndFormsByDelivery(object):
    """ A report containing all boxes, tax forms and single documents of a
    single municipality by delivery.

    """

    def __init__(self, session, start, end, type, municipality_id):
        self.session = session
        self.start = start
        self.end = end
        self.type = type
        self.municipality_id = municipality_id

    @property
    def municipality(self):
        query = self.session.query(Municipality)
        query = query.filter_by(id=self.municipality_id)
        return query.one()

    @cached_property
    def columns(self):
        return [
            'dispatch_date',
            'delivery_number',
            'return_scanned_tax_forms_older',
            'return_scanned_tax_forms_last_year',
            'return_scanned_tax_forms_current_year',
            'return_scanned_tax_forms',
            'return_scanned_single_documents',
            'return_boxes'

        ]

    def query(self):
        query = self.session.query(
            *[getattr(ScanJob, column) for column in self.columns]
        )
        query = query.filter(
            ScanJob.dispatch_date >= self.start,
            ScanJob.dispatch_date <= self.end,
            ScanJob.municipality_id == self.municipality_id
        )
        if self.type in ('normal', 'express'):
            query = query.filter(ScanJob.type == self.type)
        query = query.order_by(
            ScanJob.dispatch_date,
            ScanJob.delivery_number
        )
        return query

    def total(self):
        subquery = self.query().subquery()
        query = self.session.query(
            zero('dispatch_date'),
            zero('delivery_number'),
            *[sum(subquery.c, column) for column in self.columns[2:]],
        )
        return query.one()



from morepath import redirect
from onegov.user import User
from onegov.user import UserCollection
from onegov.wtfs import _
from onegov.wtfs import WtfsApp
from onegov.wtfs.forms import UnrestrictedUserForm
from onegov.wtfs.forms import UserForm
from onegov.wtfs.layouts import AddUserLayout
from onegov.wtfs.layouts import EditUserLayout
from onegov.wtfs.layouts import UserLayout
from onegov.wtfs.layouts import UsersLayout
from onegov.wtfs.security import AddModel
from onegov.wtfs.security import AddModelUnrestricted
from onegov.wtfs.security import DeleteModel
from onegov.wtfs.security import EditModel
from onegov.wtfs.security import EditModelUnrestricted
from onegov.wtfs.security import ViewModel


@WtfsApp.html(
    model=UserCollection,
    template='users.pt',
    permission=ViewModel
)
def view_users(self, request):
    """ View the list of users. """
    layout = UsersLayout(self, request)

    return {
        'layout': layout,
        'permission': ViewModel
    }


@WtfsApp.form(
    model=UserCollection,
    name='add-unrestricted',
    template='form.pt',
    permission=AddModelUnrestricted,
    form=UnrestrictedUserForm
)
def add_user_unrestricted(self, request, form):
    """ Create a new user. """
    layout = AddUserLayout(self, request)

    if form.submitted(request):
        user = User()
        form.update_model(user)
        request.session.add(user)
        request.message(_("User added."), 'success')
        return redirect(layout.success_url)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url
    }


@WtfsApp.form(
    model=UserCollection,
    name='add',
    template='form.pt',
    permission=AddModel,
    form=UserForm
)
def add_user(self, request, form):
    """ Create a new user. """

    if request.has_permission(self, AddModelUnrestricted):
        return redirect(request.link(self, name='add-unrestricted'))

    layout = AddUserLayout(self, request)

    if form.submitted(request):
        user = User()
        form.update_model(user)
        request.session.add(user)
        request.message(_("User added."), 'success')
        return redirect(layout.success_url)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url
    }


@WtfsApp.html(
    model=User,
    template='user.pt',
    permission=ViewModel
)
def view_user(self, request):
    """ View a single user. """
    layout = UserLayout(self, request)

    return {
        'layout': layout,
    }


@WtfsApp.form(
    model=User,
    name='edit-unrestricted',
    template='form.pt',
    permission=EditModelUnrestricted,
    form=UnrestrictedUserForm
)
def edit_user_unrestricted(self, request, form):
    """ Edit a user. """

    layout = EditUserLayout(self, request)

    if form.submitted(request):
        form.update_model(self)
        request.message(_("User modified."), 'success')
        return redirect(layout.success_url)

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url,
    }


@WtfsApp.form(
    model=User,
    name='edit',
    template='form.pt',
    permission=EditModel,
    form=UserForm
)
def edit_user(self, request, form):
    """ Edit a user. """

    if request.has_permission(self, EditModelUnrestricted):
        return redirect(request.link(self, name='edit-unrestricted'))

    layout = EditUserLayout(self, request)

    if form.submitted(request):
        form.update_model(self)
        request.message(_("User modified."), 'success')
        return redirect(layout.success_url)

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url,
    }


@WtfsApp.view(
    model=User,
    request_method='DELETE',
    permission=DeleteModel
)
def delete_user(self, request):
    """ Delete a user. """

    request.assert_valid_csrf_token()
    self.logout_all_sessions(request)
    UserCollection(request.session).delete(self.username)
    request.message(_("User deleted."), 'success')



from onegov.core.crypto import random_password
from onegov.core.orm.func import unaccent
from onegov.form import Form
from onegov.form.fields import ChosenSelectField
from onegov.form.validators import InputRequiredIf
from onegov.form.validators import UniqueColumnValue
from onegov.user import User
from onegov.wtfs import _
from onegov.wtfs.models import Municipality
from wtforms import BooleanField
from wtforms import RadioField
from wtforms import StringField
from wtforms.validators import Email
from wtforms.validators import InputRequired


class UserForm(Form):

    realname = StringField(
        label=_("Name"),
        validators=[
            InputRequired()
        ]
    )

    username = StringField(
        label=_("E-Mail"),
        validators=[
            InputRequired(),
            Email(),
            UniqueColumnValue(User)
        ]
    )

    contact = BooleanField(
        label=_("Contact"),
        depends_on=('role', '!admin'),
    )

    def update_model(self, model):
        model.realname = self.realname.data
        model.username = self.username.data
        model.data = model.data or {}
        model.data['contact'] = self.contact.data or False
        model.role = 'member'
        model.group_id = self.request.identity.groupid

        # set some initial values if we create this user
        if not model.password:
            model.password = model.password or random_password(16)
        if not model.modified:
            model.modified = model.timestamp()

        model.logout_all_sessions(self.request)

    def apply_model(self, model):
        self.realname.data = model.realname
        self.username.data = model.username
        self.contact.data = (model.data or {}).get('contact', False)


class UnrestrictedUserForm(UserForm):

    role = RadioField(
        label=_("Role"),
        choices=[
            ('admin', _("Admin")),
            ('editor', _("Editor")),
            ('member', _("Member"))
        ],
        default='member',
        validators=[
            InputRequired()
        ]
    )

    municipality_id = ChosenSelectField(
        label=_("Municipality"),
        choices=[],
        validators=[
            InputRequiredIf('role', 'editor')
        ],
        depends_on=('role', '!admin'),
    )

    def on_request(self):
        query = self.request.session.query(
            Municipality.id.label('id'),
            Municipality.name.label('name'),
            Municipality.meta['bfs_number'].label('bfs_number')
        )
        query = query.order_by(unaccent(Municipality.name))
        self.municipality_id.choices = [
            (r.id.hex, f"{r.name} ({r.bfs_number})") for r in query
        ]
        self.municipality_id.choices.insert(
            0, ('', self.request.translate(_("- none -")))
        )

    def update_model(self, model):
        super().update_model(model)
        model.role = self.role.data
        model.group_id = self.municipality_id.data or None

    def apply_model(self, model):
        super().apply_model(model)
        self.role.data = model.role
        self.municipality_id.data = (
            model.group_id.hex if model.group_id else ''
        )



from onegov.form import Form
from onegov.wtfs import _
from wtforms import StringField
from wtforms import TextAreaField
from wtforms.validators import InputRequired


class NotificationForm(Form):

    title = StringField(
        label=_("Title"),
        validators=[
            InputRequired()
        ]
    )

    text = TextAreaField(
        label=_("Text"),
        validators=[
            InputRequired()
        ]
    )

    def update_model(self, model):
        model.title = self.title.data
        model.text = self.text.data

    def apply_model(self, model):
        self.title.data = model.title
        self.text.data = model.text



from dateutil.parser import parse
from onegov.wtfs import _
from onegov.wtfs.fields.csv import CsvUploadField


class MunicipalityDataUploadField(CsvUploadField):
    """ An upload field containg municipality data. """

    def __init__(self, *args, **kwargs):
        kwargs.pop('expected_headers', None)
        kwargs.pop('rename_duplicate_column_names', None)
        super().__init__(
            *args,
            **kwargs,
            expected_headers=[
                'Gemeinde-Nr.',
                'Vordefinierte Termine',
            ],
            rename_duplicate_column_names=True,
        )

    def post_validate(self, form, validation_stopped):
        super().post_validate(form, validation_stopped)
        if validation_stopped:
            return

        date_columns = [
            self.data.as_valid_identifier(name)
            for name in self.data.headers.keys() if 'definiert' in name
        ]

        errors = []
        data = {}
        for line in self.data.lines:
            try:
                bfs_number = int(line.gemeinde_nr_)
                dates = [getattr(line, column) for column in date_columns]
                dates = [parse(d, dayfirst=True).date() for d in dates if d]
            except (AssertionError, TypeError, ValueError):
                errors.append(line.rownumber)
            else:
                data[bfs_number] = {'dates': dates}

        if errors:
            raise ValueError(_(
                "Some rows contain invalid values: ${errors}.",
                mapping={'errors': ', '.join((str(e) for e in errors))}
            ))

        self.data = data




def test_view_exceptions(client):
    assert (
        "Sie versuchen eine Seite zu öffnen, für die Sie nicht autorisiert "
        "sind"
    ) in client.get('/municipalities/add', status=403)

    assert (
        "Die angeforderte Seite konnte nicht gefunden werden."
    ) in client.get('/abstimmungen', status=404)



from morepath import redirect
from onegov.wtfs import _
from onegov.wtfs import WtfsApp
from onegov.wtfs.forms import ReportSelectionForm
from onegov.wtfs.layouts import ReportBoxesAndFormsByDeliveryLayout
from onegov.wtfs.layouts import ReportBoxesAndFormsLayout
from onegov.wtfs.layouts import ReportBoxesLayout
from onegov.wtfs.layouts import ReportFormsByMunicipalityLayout
from onegov.wtfs.layouts import ReportLayout
from onegov.wtfs.models import Report
from onegov.wtfs.models import ReportBoxes
from onegov.wtfs.models import ReportBoxesAndForms
from onegov.wtfs.models import ReportBoxesAndFormsByDelivery
from onegov.wtfs.models import ReportFormsByMunicipality
from onegov.wtfs.security import ViewModel


@WtfsApp.form(
    model=Report,
    template='form.pt',
    permission=ViewModel,
    form=ReportSelectionForm
)
def view_select_report(self, request, form):
    layout = ReportLayout(self, request)

    if form.submitted(request):
        return redirect(request.link(form.get_model()))

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Show")
    }


@WtfsApp.html(
    model=ReportBoxes,
    template='report_boxes.pt',
    permission=ViewModel
)
def view_report_boxes(self, request):
    return {'layout': ReportBoxesLayout(self, request)}


@WtfsApp.html(
    model=ReportBoxesAndForms,
    template='report_boxes_and_forms.pt',
    permission=ViewModel
)
def view_report_boxes_and_forms(self, request):
    return {'layout': ReportBoxesAndFormsLayout(self, request)}


@WtfsApp.html(
    model=ReportFormsByMunicipality,
    template='report_forms.pt',
    permission=ViewModel
)
def view_report_forms(self, request):
    return {'layout': ReportFormsByMunicipalityLayout(self, request)}


@WtfsApp.html(
    model=ReportBoxesAndFormsByDelivery,
    template='report_delivery.pt',
    permission=ViewModel
)
def view_report_delivery(self, request):
    return {'layout': ReportBoxesAndFormsByDeliveryLayout(self, request)}



from onegov.core.utils import module_path
from onegov.foundation import BaseTheme


class WtfsTheme(BaseTheme):
    name = 'onegov.wtfs.foundation'
    version = '1.1.2'

    @property
    def pre_imports(self):
        return ['wtfs-foundation-mods']

    @property
    def post_imports(self):
        return super().post_imports + [
            'mixin',
            'header',
            'footer',
            'form',
            'chosen',
            'table',
            'alert',
            'wtfs'
        ]

    @property
    def extra_search_paths(self):
        base_paths = super().extra_search_paths
        return [module_path('onegov.wtfs.theme', 'styles')] + base_paths



from onegov.core.orm.mixins import meta_property
from onegov.core.orm.mixins import TimestampMixin
from onegov.user import UserGroup
from onegov.wtfs.models.payment_type import PaymentType
from sqlalchemy.orm import object_session


class Municipality(UserGroup, TimestampMixin):
    """ A municipality / user group. """

    __mapper_args__ = {'polymorphic_identity': 'wtfs'}

    #: The name of the municipality.
    bfs_number = meta_property()

    #: The address supplement, used for invoices.
    address_supplement = meta_property()

    #: The GPN number, used for invoices.
    gpn_number = meta_property()

    #: The payment type. Typically normal (7.00) or special (8.50).
    payment_type = meta_property('payment_type')

    @property
    def price_per_quantity(self):
        if self.payment_type:
            query = object_session(self).query(PaymentType)
            query = query.filter_by(name=self.payment_type)
            payment_type = query.first()
            if payment_type:
                return payment_type.price_per_quantity or 0

        return 0

    @property
    def has_data(self):
        if self.pickup_dates.first() or self.scan_jobs.first():
            return True
        return False



from freezegun import freeze_time
from onegov.core.request import CoreRequest
from unittest.mock import patch
from webtest.forms import Upload


def test_views_invoices(client):
    # Add a municipality with dates
    client.login_admin()

    add = client.get('/municipalities').click(href='/add')
    add.form['name'] = "Adlikon"
    add.form['bfs_number'] = '1'
    add.form['gpn_number'] = '11223344'
    add.form['payment_type'] = 'normal'
    assert "Adlikon" in add.form.submit().follow()

    upload = client.get('/municipalities').click("Daten importieren")
    upload.form['file'] = Upload(
        'test.csv',
        "Gemeinde-Nr,Vordefinierte Termine\n1,5.1.2019".encode('utf-8'),
        'text/csv'
    )
    assert "Gemeindedaten importiert." in upload.form.submit().follow()

    # Add a scan job
    with freeze_time("2019-01-01"):
        add = client.get('/scan-jobs/unrestricted').click(href='/add')
        add.form['type'].select("normal")
        add.form['municipality_id'].select(text="Adlikon (1)")
        add.form['dispatch_date'] = "2019-01-05"
        add.form['dispatch_tax_forms_older'] = "30"
        add.form['dispatch_tax_forms_last_year'] = "20"
        add.form['dispatch_tax_forms_current_year'] = "10"
        assert "Scan-Auftrag hinzugefügt." in add.form.submit().maybe_follow()

        edit = client.get('/scan-jobs/unrestricted')\
            .click("05.01.2019").click("Bearbeiten")
        edit.form['return_date'] = "2019-01-10"
        edit.form['return_unscanned_tax_forms_older'] = "3"
        edit.form['return_unscanned_tax_forms_last_year'] = "2"
        edit.form['return_unscanned_tax_forms_current_year'] = "1"

        assert "Scan-Auftrag geändert." in edit.form.submit().maybe_follow()

    with freeze_time("2019-01-02"):
        create = client.get('/invoice')
        create.form['from_date'] = "2019-01-01"
        create.form['to_date'] = "2019-01-07"
        create.form['cs2_user'] = "123"
        create.form['subject'] = "Rechnungen"
        create.form['municipality_id'].select(text="Adlikon (1)")
        create.form['accounting_unit'] = "456"
        create.form['revenue_account'] = "789"
        created = create.form.submit()
        assert created.headers['Content-Type'] == 'text/csv; charset=UTF-8'
        assert created.headers['Content-Disposition'] == (
            'inline; filename=rechnungen.csv'
        )

        invoice = created.text
        assert "11223344" in invoice
        assert "Rechnungen" in invoice
        assert "123" in invoice
        assert "456" in invoice
        assert "789" in invoice
        assert "02.01.2019" in invoice
        assert "54000" in invoice
        assert "70000000" in invoice


@patch.object(CoreRequest, 'assert_valid_csrf_token')
def test_views_invoice_permissions(mock_method, client):
    client.get('/invoice', status=403)

    client.login_member()
    client.get('/invoice', status=403)
    client.logout()

    client.login_editor()
    client.get('/invoice', status=403)
    client.logout()

    client.login_admin()
    client.get('/invoice')
    client.logout()



from morepath import redirect
from onegov.wtfs import _
from onegov.wtfs import WtfsApp
from onegov.wtfs.collections import PaymentTypeCollection
from onegov.wtfs.forms import PaymentTypesForm
from onegov.wtfs.layouts import PaymentTypesLayout
from onegov.wtfs.security import EditModel


@WtfsApp.form(
    model=PaymentTypeCollection,
    template='form.pt',
    permission=EditModel,
    form=PaymentTypesForm.get_form_class
)
def manage_payments(self, request, form):
    """ Manage payment types. """

    layout = PaymentTypesLayout(self, request)

    if form.submitted(request):
        form.update_model(self)
        request.message(_("Payment types modified."), 'success')
        return redirect(layout.success_url)

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': layout,
        'form': form,
        'button_text': _("Save"),
        'cancel': layout.cancel_url
    }



from onegov.wtfs.models.daily_list import DailyList
from onegov.wtfs.models.daily_list import DailyListBoxes
from onegov.wtfs.models.daily_list import DailyListBoxesAndForms
from onegov.wtfs.models.invoice import Invoice
from onegov.wtfs.models.municipality import Municipality
from onegov.wtfs.models.notification import Notification
from onegov.wtfs.models.payment_type import PaymentType
from onegov.wtfs.models.pickup_date import PickupDate
from onegov.wtfs.models.principal import Principal
from onegov.wtfs.models.report import Report
from onegov.wtfs.models.report import ReportBoxes
from onegov.wtfs.models.report import ReportBoxesAndForms
from onegov.wtfs.models.report import ReportBoxesAndFormsByDelivery
from onegov.wtfs.models.report import ReportFormsByMunicipality
from onegov.wtfs.models.scan_job import ScanJob


__all__ = (
    'DailyList',
    'DailyListBoxes',
    'DailyListBoxesAndForms',
    'Invoice',
    'Municipality',
    'Notification',
    'PaymentType',
    'PickupDate',
    'Principal',
    'Report',
    'ReportBoxes',
    'ReportBoxesAndForms',
    'ReportBoxesAndFormsByDelivery',
    'ReportFormsByMunicipality',
    'ScanJob',
)



from onegov.core.request import CoreRequest
from unittest.mock import patch


def test_views_user(client):
    client.login_admin()

    # Add a municipality
    add = client.get('/municipalities').click(href='/add')
    add.form['name'] = "Adlikon"
    add.form['bfs_number'] = '1'
    add.form['payment_type'] = 'normal'
    assert "Adlikon" in add.form.submit().follow()

    # Add a user
    add = client.get('/users').click(href='/add')
    add.form['realname'] = "Hans Muster"
    add.form['username'] = "hans.muster@winterthur.ch"
    add.form['role'].select('editor')
    add.form['municipality_id'].select(text="Adlikon (1)")
    add.form['contact'] = True
    added = add.form.submit().follow()
    assert "Benutzer hinzugefügt." in added
    assert "Hans Muster" in added
    assert "Hans Muster" in client.get('/municipalities').click("Adlikon")

    # Add a user without a group
    add = client.get('/users').click(href='/add')
    add.form['realname'] = "Optimo X"
    add.form['username'] = "info@optimo.info"
    add.form['role'].select('member')
    added = add.form.submit().follow()
    assert "Benutzer hinzugefügt." in added
    assert "Optimo X" in added

    # View users
    view = client.get('/users').click("Hans Muster")
    assert "Gemeindeadministrator" in view
    assert "hans.muster@winterthur.ch" in view
    assert "✔︎" in view
    assert "Adlikon" in view

    view = client.get('/users').click("Optimo X")
    assert "Benutzer" in view
    assert "info@optimo.info" in view
    assert "✘︎" in view

    # Edit user
    edit = client.get('/users').click("Hans Muster").click("Bearbeiten")
    edit.form['realname'] = "Hans-Peter Muster"
    edit.form['username'] = "hans-peter.muster@winterthur.ch"
    edit.form['role'].select('member')
    assert "Benutzer geändert." in edit.form.submit().follow()
    view = client.get('/users').click("Hans-Peter Muster")
    assert "Hans-Peter Muster" in view
    assert "Benutzer" in view
    assert "hans-peter.muster@winterthur.ch" in view
    assert "✔︎" in view
    assert "Adlikon" in view

    # Delete user
    deleted = client.get('/users').click("Hans-Peter Muster").click("Löschen")
    assert deleted.status_int == 200
    deleted = client.get('/users').click("Optimo X").click("Löschen")
    assert deleted.status_int == 200
    assert "Hans-Peter Muster" not in client.get('/users')
    assert "Optimo X" not in client.get('/users')


def test_views_user_editor(client):
    client.login_editor()

    # Add a user
    add = client.get('/users').click(href='/add')
    add.form['realname'] = "Hans Muster"
    add.form['username'] = "hans.muster@winterthur.ch"
    add.form['contact'] = True
    added = add.form.submit().follow()
    assert "Benutzer hinzugefügt." in added
    assert "Hans Muster" in added

    # View the user
    view = client.get('/users').click("Hans Muster")
    assert "Benutzer" in view
    assert "hans.muster@winterthur.ch" in view
    assert "✔︎" in view
    assert "My Municipality" in view

    # Edit the user
    edit = client.get('/users').click("Hans Muster").click("Bearbeiten")
    edit.form['realname'] = "Hans-Peter Muster"
    edit.form['username'] = "hans-peter.muster@winterthur.ch"
    edit.form['contact'] = False
    assert "Benutzer geändert." in edit.form.submit().follow()
    view = client.get('/users').click("Hans-Peter Muster")
    assert "Hans-Peter Muster" in view
    assert "Benutzer" in view
    assert "hans-peter.muster@winterthur.ch" in view
    assert "✘︎" in view
    assert "My Municipality" in view

    # Delete the user
    deleted = client.get('/users').click("Hans-Peter Muster").click("Löschen")
    assert deleted.status_int == 200
    assert "Hans-Peter Muster" not in client.get('/users')


@patch.object(CoreRequest, 'assert_valid_csrf_token')
def test_views_users_permissions(mock_method, client):
    client.login_admin()

    add = client.get('/users').click(href='/add')
    add.form['realname'] = "Optimo X"
    add.form['username'] = "info@optimo.info"
    add.form['role'].select('member')
    assert "Benutzer hinzugefügt." in add.form.submit().follow()

    client.logout()

    urls = [
        '/users',
        '/users/add',
        '/users/add-unrestricted'
    ]
    users = [
        'member%40example.org',
        'editor%40example.org',
        'info%40optimo.info',
        'admin%40example.org',
    ]

    for url in urls:
        client.get(url, status=403)
    for user in users:
        client.get(f'/user/{user}', status=403)
        client.get(f'/user/{user}/edit', status=403)
        client.get(f'/user/{user}/edit-unrestricted', status=403)
        client.delete(f'/user/{user}', status=403)

    client.login_member()
    for url in urls:
        client.get(url, status=403)
    for user in users:
        client.get(f'/user/{user}', status=403)
        client.get(f'/user/{user}/edit', status=403)
        client.get(f'/user/{user}/edit-unrestricted', status=403)
        client.delete(f'/user/{user}', status=403)
    client.logout()

    client.login_editor()
    client.get('/users')
    client.get('/users/add')
    client.get('/users/add-unrestricted', status=403)
    for user in users[:1]:
        client.get(f'/user/{user}')
        client.get(f'/user/{user}/edit')
        client.get(f'/user/{user}/edit-unrestricted', status=403)
        client.delete(f'/user/{user}')
    for user in users[1:]:
        client.get(f'/user/{user}', status=403)
        client.get(f'/user/{user}/edit', status=403)
        client.get(f'/user/{user}/edit-unrestricted', status=403)
        client.delete(f'/user/{user}', status=403)
    client.logout()

    client.login_admin()
    for user in users[1:]:
        client.get(f'/user/{user}')
        client.get(f'/user/{user}/edit')
        client.get(f'/user/{user}/edit-unrestricted')
    for user in users[1:-1]:
        client.delete(f'/user/{user}')
    for user in users[-1:]:
        client.delete(f'/user/{user}', status=403)
    client.logout()



from lxml.html import document_fromstring
from onegov_testing import utils
from webtest import TestApp as Client
import onegov


def test_view_permissions():
    utils.assert_explicit_permissions(
        onegov.wtfs,
        onegov.wtfs.WtfsApp
    )


def test_view_login_logout(client):
    error = "Unbekannter Benutzername oder falsches Passwort"

    for user in ('admin', 'editor', 'member'):
        login = client.get('/').maybe_follow().click('Anmelden')
        login.form['username'] = f'{user}@example.org'
        login.form['password'] = 'hunter1'
        page = login.form.submit().maybe_follow()
        assert error in page
        assert 'Abmelden' not in page
        assert 'Anmelden' in page

        login = client.get('/').maybe_follow().click('Anmelden')
        login.form['username'] = f'{user}@example.org'
        login.form['password'] = 'hunter2'
        page = login.form.submit().maybe_follow()
        assert error not in page
        assert 'Abmelden' in page
        assert 'Anmelden' not in page

        page = page.click('Abmelden').maybe_follow()
        assert 'Abmelden' not in page
        assert 'Anmelden' in page


def test_view_reset_password(wtfs_app):
    client = Client(wtfs_app)

    home = client.get('/').maybe_follow()

    request_page = home.click('Anmelden').click('Passwort zurücksetzen')
    request_page.form['email'] = 'someone@example.org'
    request_page.form.submit()
    assert len(wtfs_app.smtp.outbox) == 0

    request_page.form['email'] = 'admin@example.org'
    request_page.form.submit()
    assert len(wtfs_app.smtp.outbox) == 1

    message = wtfs_app.smtp.outbox[0]
    message = message.get_payload(1).get_payload(decode=True)
    message = message.decode('iso-8859-1')
    link = list(document_fromstring(message).iterlinks())[0][2]
    token = link.split('token=')[1]

    reset_page = client.get(link)
    assert token in reset_page.text

    reset_page.form['email'] = 'someone_else@example.org'
    reset_page.form['password'] = 'password1'
    reset_page = reset_page.form.submit()
    assert "Ungültige E-Mail oder abgelaufener Link" in reset_page
    assert token in reset_page.text

    reset_page.form['email'] = 'admin@example.org'
    reset_page.form['password'] = '1234'
    reset_page = reset_page.form.submit()
    assert "Feld muss mindestens 8 Zeichen beinhalten" in reset_page
    assert token in reset_page.text

    reset_page.form['email'] = 'admin@example.org'
    reset_page.form['password'] = 'password2'
    assert "Passwort geändert" in reset_page.form.submit().maybe_follow()

    reset_page.form['email'] = 'admin@example.org'
    reset_page.form['password'] = 'password3'
    reset_page = reset_page.form.submit()
    assert "Ungültige E-Mail oder abgelaufener Link" in reset_page

    login_page = client.get('/auth/login')
    login_page.form['username'] = 'admin@example.org'
    login_page.form['password'] = 'hunter2'
    login_page = login_page.form.submit()
    assert "Unbekannter Benutzername oder falsches Passwort" in login_page

    login_page.form['username'] = 'admin@example.org'
    login_page.form['password'] = 'password2'
    login_page = login_page.form.submit().maybe_follow()
    assert "Abmelden" in login_page.maybe_follow()




def test_view_home(client):
    assert 'auth/login' in client.get('/').follow().request.url

    client.login_admin()

    assert "/scan-jobs" in client.get('/').follow().request.url

    add = client.get('/notifications').click(href='/add')
    add.form['title'] = "Systemunterbruch"
    add.form['text'] = "Am 23. Februar 2017, finden Wartungsarbeiten statt."
    assert "Benachrichtigung hinzugefügt." in add.form.submit().follow()

    assert "Systemunterbruch" in client.get('/')
    assert "Wartungsarbeiten" in client.get('/')

    client.logout()



from cached_property import cached_property
from onegov.core.elements import Link
from onegov.wtfs import _
from onegov.wtfs.collections import PaymentTypeCollection
from onegov.wtfs.layouts.default import DefaultLayout
from onegov.wtfs.security import EditModel


class InvoiceLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Create invoice")

    @cached_property
    def editbar_links(self):
        result = []
        model = PaymentTypeCollection(self.request.session)
        if self.request.has_permission(model, EditModel):
            result.append(
                Link(
                    text=_("Manage payment types"),
                    url=self.request.link(model),
                    attrs={'class': 'payment-icon'}
                )
            )
        return result

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(self.title, self.request.link(self.model))
        ]

    @cached_property
    def cancel_url(self):
        return self.invoices_url

    @cached_property
    def success_url(self):
        return self.invoices_url



from onegov.wtfs.collections.municipalities import MunicipalityCollection
from onegov.wtfs.collections.notifications import NotificationCollection
from onegov.wtfs.collections.payment_types import PaymentTypeCollection
from onegov.wtfs.collections.scan_jobs import ScanJobCollection


__all__ = (
    'MunicipalityCollection',
    'NotificationCollection',
    'PaymentTypeCollection',
    'ScanJobCollection',
)



import logging

from translationstring import TranslationStringFactory

log = logging.getLogger('onegov.wtfs')
log.addHandler(logging.NullHandler())
_ = TranslationStringFactory('onegov.wtfs')

from onegov.wtfs.app import WtfsApp  # noqa

__all__ = (
    '_',
    'log',
    'WtfsApp'
)



from cached_property import cached_property
from onegov.core.elements import Link
from onegov.wtfs import _
from onegov.wtfs.layouts.default import DefaultLayout


class DailyListLayout(DefaultLayout):

    @cached_property
    def title(self):
        return _("Daily list")

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(self.title, self.request.link(self.model)),
        ]


class SpecificDailyListBaseLayout(DefaultLayout):

    date_custom_format = 'EEEE dd. MMMM yyyy'

    @cached_property
    def subtitle(self):
        return self.format_date(self.model.date, 'date_custom')

    @cached_property
    def editbar_links(self):
        result = []
        result.append(
            Link(
                text=_("Print"),
                url='#',
                attrs={
                    'class': 'print-icon',
                    'onclick': 'window.print();return false;'
                }
            )
        )
        return result

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Daily list"), self.daily_list_url),
            Link(self.title, "#"),
            Link(self.subtitle, self.request.link(self.model)),
        ]


class DailyListBoxesLayout(SpecificDailyListBaseLayout):

    @cached_property
    def title(self):
        return _("Boxes")


class DailyListBoxesAndFormsLayout(SpecificDailyListBaseLayout):

    @cached_property
    def title(self):
        return _("Boxes and forms")
